import os
from dataclasses import dataclass
from typing import Union, TextIO, Optional, Set, List, Any, Callable, Dict, Tuple
import logging

import click
from jinja2 import Template

from linkml_runtime.linkml_model.meta import SchemaDefinition, ClassDefinition, SlotDefinition, Element, \
    ClassDefinitionName, \
    SlotDefinitionName, \
    ElementName, TypeDefinitionName, TypeDefinition, Definition, DefinitionName, PermissibleValue
from linkml.utils.generator import Generator, shared_arguments
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.schemaview import SchemaView


macros = """
{% macro slot(s, c=None) -%}
// MACRO SLOT: {{s.name}} {{c}}

{%- endmacro %}
"""

template = macros + """


/**
 Schema: {{schema.name}}
*/

// Declarations
#define RDF_TYPE "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"

.type identifier = symbol
.type value = symbol

// Mapping from RDF
.decl triple(s:symbol, p:symbol, o:symbol)
.input triple
.decl literal_number(s:symbol, o:number)
.input literal_number
.decl literal_symbol(s:symbol, o:symbol)
.input literal_symbol

// closure
.decl uri_subsumed_by(s:symbol, o:symbol)
uri_subsumed_by(s,o) :- uri_subsumed_by(s,z), uri_subsumed_by(z,o).

.decl validation_result(type: symbol, subject: symbol, instantiates: symbol, path: symbol, value: symbol, info:symbol)
.output validation_result


{% if 'datalog' in schemaview.schema.annotations %}
// ------------------
// -- SCHEMA RULES --
// ------------------
{{ schemaview.schema.annotations['datalog'].value }}
{% endif %}

// -------------
// -- Slots --
// -------------
{% for s in schemaview.all_slots().values() %}
{% set dltype = gen.datalog_type(s) %}
// Slot: {{s.name}} TYPE: {{ dltype }}
{{slot(s)}}
{% set spred = gen.pred(s) -%}
.decl {{ spred }}_asserted(i: identifier, v: {{ dltype }})
.decl {{ spred }}(i: identifier, v: {{ dltype }})
.output {{ spred }}
{{ spred }}(i, v) :- 
    {{ spred }}_asserted(i, v).
{{ spred }}_asserted(i, v) :- 
    {% if dltype == 'identifier' %}
    triple(i, "{{ gen.uri(s) }}", v).
    {% else %}
    triple(i, "{{ gen.uri(s) }}", x),
    {% if dltype == 'number' %}
    literal_number(x, v).
    {% else %}
     literal_symbol(x, v).
    {% endif %}
    {% endif %}
    
uri_subsumed_by("{{ gen.uri(s) }}", "{{ gen.uri(s) }}").
{% for p in gen.parents(s) %}
{{ gen.pred(p) }}(i, v) :- {{ spred }}(i, v).
uri_subsumed_by("{{ gen.uri(s) }}", "{{ gen.uri(p) }}").
{% endfor %}

{% if s.inverse %}
// inverse
{{ spred }}(i, v) :- {{ gen.pred(s.inverse) }}(v, i). 
{% endif %}

{% if s.symmetric %}
// symmetric
{{ spred }}(i, v) :- {{ spred }}(v, i). 
{% endif %}

{% if gen.is_transitive(s) %}
// transitive
{{ spred }}(i, v) :- 
    {{ spred }}(i, z),
    {{ spred }}(z, v).
{% endif %}

{% if 'transitive_closure_of' in s.annotations %}
// transitive
{{ spred }}(i, v) :- 
    {{ gen.pred(s.annotations['transitive_closure_of'].value) }}(i, v).
{% endif %}

// DOMAIN AND RANGE

validation_result(
  "sh:ClosedConstraintComponent",
  i,
  "{{s.name}}",
  "{{s.name}}",
  //to_string(v),
  "",
  "Expected domain is {{gen.domains(s)}}") :-
    {{spred}}(i, _v)
    {%- for domain in gen.domains(s) -%}
    , ! {{ gen.pred(domain) }}(i)
    {%- endfor %} .
    
{% if s.range and not s.range in schemaview.all_types() %}
validation_result(
  "sh:Range",
  i,
  "{{s.name}}",
  "{{s.name}}",
  //as(v, value),
  "",
  "Expected range is {{s.range}}") :-
    {{spred}}(i, v),
    ! {{ gen.pred(s.range) }}(v).
{% endif %}
    
{% endfor %}
// end of slots block

// -------------
// -- CLASSES --
// -------------
{% for c in schemaview.all_classes().values() %}
// Class: {{c.name}}
{% set cpred = gen.pred(c) -%}
.decl {{ cpred }}(i: symbol)
.decl {{ cpred }}_asserted(i: identifier)
.output {{ cpred }}
{{ cpred }}_asserted(i) :- triple(i, RDF_TYPE, "{{ gen.uri(c) }}").
{{ cpred }}(i) :- {{ cpred }}_asserted(i).
{% if c.is_a %}
{{ gen.pred(c.is_a) }}(i) :- {{ cpred }}(i).
{% endif %}
{% for p in c.mixins %}
{{ gen.pred(p) }}(i) :- {{ cpred }}(i).
{% endfor %}


{% if gen.reification_of(c.name) %}
{% set reif = gen.reification_of(c.name) %}
// DE-REIFICATION RULE:
triple(i, p, v) :-
        {% if reif.subject %}
        {{gen.pred(reif.subject)}}(r, i)
        {% else %}
        triple(i, _container_prop, r),
        {% endif %}
        {{gen.pred(reif.object)}}(r, v),
        {{gen.pred(reif.predicate)}}(r, p).
{% endif %} 

{% if c.defining_slots %}
// Auto-classification from defining slots: {{c.defining_slots}}

{{ cpred }}(i) :-
    {{ gen.pred(c.is_a) }}(i)
    {% for ds in c.defining_slots %}
    {% set islot = schemaview.induced_slot(ds, c.name) %}
    {% set spred = gen.pred(islot) %}
    {% if islot.subproperty_of %}
    , {{ spred }}(i, v_{{ spred }}), uri_subsumed_by(v_{{ spred }}, "{{ gen.uri(islot.subproperty_of) }}")
    {% else %}
    , {{ spred }}(i, v_{{ spred }}), {{ gen.pred(islot.range) }}(v_{{ spred }})
    {% endif %}
    {% endfor %}
    .
{% endif %}

{% for s in schemaview.class_induced_slots(c.name) %}
{% set spred = gen.class_slot_pred(c, s) -%}
{% set dltype = gen.datalog_type(s) %}
// CLASS_SLOT {{s.name}} TYPE: {{ dltype }}
.decl {{ spred }}_asserted(i: identifier, v: {{ dltype }})
.decl {{ spred }}(i: identifier, v: {{ dltype }})
.output {{ spred }}
.output {{ spred }}_asserted
{{ spred }}(i, v) :- 
    {{ spred }}_asserted(i, v).
{{ spred }}_asserted(i, v) :- 
    {{ cpred }}(i),
    {{ gen.pred(s) }}(i, v).
// TODO: inferring default values

{% if s.inverse and False %}
{% set inv = s.inverse %}
{% set inv_range = gen.get_slot_range(inv) %}
// Inverse of: {{inv}} range: {{inv_range}}
{% set spred_inv = gen.class_slot_pred(inv_range, inv) %}
{{ spred }}(i, v) :- {{ spred_inv }}(v, i). 
{% endif %}

{% if 'classified_from' in s.annotations %}
{% set classified_from = s.annotations['classified_from'].value %}
{% set enum = schemaview.get_enum(s.range) %}
// CLASSIFYING CATEGORY FROM OTHER SLOT {{enum.name}} . {{classified_from}}
{% for pv in enum.permissible_values.values() %}
// PV = {{pv.text}}
{% if 'expr' in pv.annotations %}
{% set expr = pv.annotations['expr'] %}
{{ spred }}(i, "{{ gen.uri(pv) }}" ) :-
     {{ classified_from }}(i, v),
     {{expr.value}} .
{% endif %}
{% endfor %}
{% endif %}

{% if not s.multivalued %}
validation_result(
  "sh:MaxCountConstraintComponent",
  i,
  "{{ cpred }}",
  "{{ s.name }}",
  "",
  //v1,
  "got two distinct values for subject and predicate") :-
    {{ spred }}(i, v1),
    {{ spred }}(i, v2),
    v1 != v2. 
{% endif %}

{% if s.required %}
validation_result(
  "sh:MinCountConstraintComponent",
  i,
  "{{ cpred }}",
  "{{ s.name }}",
  "",
  "") :-
    {{ cpred }}(i),
    ! {{spred}}(i, _).
{% endif %}

{% if s.maximum_value %}
validation_result(
  "sh:MaxInclusiveConstraintComponent",
  i,
  "{{ cpred }}",
  "{{s.name}}",
  to_string(v),
  "Maximum is {{s.maximum_value}}") :-
    {{ cpred }}(i),
    {{spred}}(i, v),
    v > {{ s.maximum_value }}.
{% endif %}

{% if s.range and not s.range in schemaview.all_types() %}
validation_result(
  "sh:ClassConstraintComponent",
  i,
  "{{ cpred }}",
  "{{s.name}}",
  v,
  "Expected range is {{s.range}}") :-
    {{ cpred }}(i),
    {{spred}}(i, v),
    ! {{ gen.pred(s.range) }}(v).
{% endif %}

{% endfor %}
// end of class slots block



{% endfor %}
// end of classes block

// -------------
// -- Types --
// -------------
{% for t in schemaview.all_types().values() %}
{% set type_type = gen.type_to_datalog_type(t) %}
// Type: {{t.name}} . {{ type_type }}
//{% set tpred = gen.pred(t) -%}
//.decl {{ tpred }}(i: {{type_type}})
//{% if type_type == 'number' %}
//{% endif %}
//{{ tpred }}(i) :- literal_symbol(i, _).
//{{ tpred }}(i) :- literal_number(i, _).
// TODO!
{% endfor %}
// end of types block

// -------------
// -- Enums --
// -------------
{% for e in schemaview.all_enums().values() %}
// Enum: {{e.name}}
{% set epred = gen.pred(e) -%}
.decl {{ epred }}(i: symbol)
{{ epred }}(i) :- literal_symbol(i, _).
// TODO!
{% for pv in e.permissible_values.values() %}
{% if pv.meaning %}
{{ epred }}("{{ gen.meaning_uri(pv.meaning) }}").
{% else %}
{% endif %}
{% endfor %}
{% endfor %}
// end of enums block

"""

@dataclass
class Reification:
    subject: SlotDefinitionName = None
    predicate: SlotDefinitionName = None
    object: SlotDefinitionName = None

class DatalogGenerator(Generator):
    """
    Generates Souffle datalog from a LinkML schema where the domain of discourse is RDF triples

    """
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ['dl']
    visit_all_class_slots = True
    visited = set()
    type_field_uris: List[str] = []
    schemaview: SchemaView = []

    def __init__(self, schema: Union[str, TextIO, SchemaDefinition], format: str = valid_formats[0], **kwargs) -> None:
        self.format = format
        self.schemaview = SchemaView(schema)

    def serialize(self, **kwargs) -> str:
        sv = self.schemaview
        template_obj = Template(template)
        code = template_obj.render(schemaview=self.schemaview,
                                   schema=self.schemaview.schema,
                                   gen=self)
        return code


    def pred(self, el: Union[Element, ElementName]) -> str:
        sv = self.schemaview
        sv: SchemaView
        if el == '':
            return 'NONE'
        if not isinstance(el, Element):
            if sv.get_element(el) is None:
                logging.error(f'No such element: "{el}" // {type(el)}')
                return 'UNDEFINED'
            el = sv.get_element(el)
        if isinstance(el, SlotDefinition):
            pred = underscore(el.name)
        else:
            pred = camelcase(el.name)
        return pred

    def class_slot_pred(self, c: Union[ClassDefinition, ClassDefinitionName], s: Union[SlotDefinition, SlotDefinitionName]) -> str:
        return f'{self.pred(c)}_{self.pred(s)}'

    def uri(self, el: Union[Element, ElementName, PermissibleValue]) -> str:
        sv = self.schemaview
        sv: SchemaView
        if isinstance(el, PermissibleValue):
            if el.meaning:
                return self.meaning_uri(el.meaning)
        if not isinstance(el, Element):
            el = sv.get_element(el)
        if el is None:
            logging.error(f'NONE')
            return 'NONE'
        return sv.get_uri(el, expand=True)

    def meaning_uri(self, curie: str):
        return self.schemaview.expand_curie(curie)

    def domains(self, slot: SlotDefinition) -> List[ClassDefinitionName]:
        sv = self.schemaview
        domains = []
        for cn in sv.all_classes().keys():
            if slot.name in sv.class_slots(cn, direct=True):
                domains.append(cn)
        if slot.domain:
            domains.append(slot.domain)
        return domains

    def get_slot_range(self, sn: SlotDefinitionName):
        """
        TEMPORARY
        """
        sv = self.schemaview
        slot = sv.get_slot(sn)
        if slot.range:
            return slot.range
        elif slot.is_a:
            return self.get_slot_range(slot.is_a)
        else:
            raise ValueError(f'No range for slot {sn}')

    def datalog_type(self, s: Union[SlotDefinition, SlotDefinitionName]) -> str:
        sv = self.schemaview
        if not isinstance(s, SlotDefinition):
            s = sv.get_slot(s)
        if s.range in sv.all_types():
            return self.type_to_datalog_type(s.range)
        return 'identifier'

    def parents(self, e: Definition) -> List[DefinitionName]:
        sv = self.schemaview
        if isinstance(e, SlotDefinition):
            return sv.slot_parents(e.name)
        elif isinstance(e, ClassDefinition):
            return sv.class_parents(e.name)
        else:
            raise ValueError(f'Must be slot or class: {e}')

    def slot_is_numeric(self, s: Union[SlotDefinition, SlotDefinitionName]) -> bool:
        return self.datalog_type(s) == 'number'

    def is_transitive(self, s: SlotDefinition) -> bool:
        return self.has_annotation(s, 'transitive') or self.has_annotation(s, 'transitive_closure_of')

    def is_reflexive(self, s: SlotDefinition) -> bool:
        return self.has_annotation(s, 'transitive')

    def is_symmetric(self, s: SlotDefinition) -> bool:
        return s.symmetric

    def has_annotation(self, s: SlotDefinition, a: str) -> bool:
        return a in s.annotations

    def type_to_datalog_type(self, t: Union[TypeDefinition, TypeDefinitionName]) -> str:
        sv = self.schemaview
        if not isinstance(t, TypeDefinition):
            t = sv.get_type(t)
        if t.base == 'int' or t.base == 'float' or t.base == 'Decimal':
            return 'number'
        else:
            return 'symbol'

    def type_is_numeric(self, t: Union[TypeDefinition, TypeDefinitionName]) -> bool:
        return self.type_to_datalog_type(t) == 'number'

    def reification_of(self, cn: ClassDefinitionName) -> Optional[Reification]:
        """
        TODO: move to schemaview
        """
        sv = self.schemaview
        if sv.is_relationship(cn):
            islots = sv.class_induced_slots(cn)
            su = None
            pr = None
            ob = None
            for islot in islots:
                for anc in sv.class_ancestors(cn):
                    slot_uri = sv.induced_slot(islot.name, anc).slot_uri
                    if slot_uri == 'rdf:subject':
                        su = islot
                    if slot_uri == 'rdf:predicate':
                        pr = islot
                    if slot_uri == 'rdf:object':
                        ob = islot
            if pr is None:
                logging.error(f'No predicate for {cn}')
                return None
            if ob is None:
                logging.error(f'No object for {cn}')
                return None
            reif = Reification(subject=su,
                               predicate=pr,
                               object=ob)
            return reif
        else:
            return None



@shared_arguments(DatalogGenerator)
@click.command()
def cli(yamlfile, dir, **kwargs):
    """ Generate Souffle datalog from a LinkML schema """
    print(DatalogGenerator(yamlfile, **kwargs).serialize(**kwargs))


if __name__ == '__main__':
    cli()



import logging
import os
from abc import abstractmethod
from enum import Enum
from numbers import Number
from typing import Optional, Any, Dict, List, Union

from linkml_runtime.dumpers import rdflib_dumper
from rdflib import Graph, URIRef
from rdflib.term import Node, BNode, Literal, Identifier
from rdflib.namespace import RDF


from linkml_runtime.dumpers.dumper_root import Dumper
from linkml_runtime.utils.schemaview import SchemaView, ElementName, PermissibleValue, PermissibleValueText
from linkml_runtime.utils.yamlutils import YAMLRoot

class Predicate(Enum):
    triple = 'triple'
    literal_number = 'literal_number'
    literal_symbol = 'literal_symbol'

    @staticmethod
    def list() -> List[str]:
        return list(map(lambda c: c.value, Predicate))


class TupleDumper(Dumper):
    """
    Dumps LinkML instance data as TSV tuples
    """

    def dump(self, element: Union[YAMLRoot, Graph], schemaview: SchemaView = None, directory=None, **kwargs):
        if isinstance(element, Graph):
            g = element
        else:
            g = rdflib_dumper.as_rdf_graph(element, schemaview, **kwargs)
        self.graph_to_tuples(g, directory=directory)

    def dumps(self, *args, **kwargs):
        return self.dump(*args, **kwargs)

    def graph_to_tuples(self, graph: Graph, directory: str) -> None:
        file_map = {}
        for p in Predicate.list():
            file_map[p] = open(os.path.join(directory, f'{p}.facts'), 'w')

        def safe_str(v: str) -> str:
            return str(v).replace('\t', '\\t').replace('\n', '\\n')

        def as_str(v: Identifier) -> str:
            if isinstance(v, Literal):
                v = v.toPython()
                if isinstance(v, Number) and not isinstance(v, bool):
                    return str(v)
                else:
                    v = safe_str(v)
                    return f'"{v}"'
            else:
                return str(v)

        def emit(predicate: Predicate, *args):
            file_map[predicate.value].write('\t'.join([as_str(a) for a in args]))
            file_map[predicate.value].write('\n')

        for s, p, o in graph.triples((None, None, None)):
            emit(Predicate.triple, s, p, o)
            if isinstance(o, Literal):
                v = o.toPython()
                if isinstance(v, Number) and not isinstance(v, bool):
                    emit(Predicate.literal_number, as_str(o), v)
                else:
                    emit(Predicate.literal_symbol, as_str(o), safe_str(v))







import csv
import json
import os
import subprocess
from dataclasses import dataclass
from typing import Any, Dict, List, Union, Tuple

import yaml
import logging

import click
from linkml.generators.pythongen import PythonGenerator
from linkml_runtime.dumpers import yaml_dumper
from linkml_runtime.linkml_model import SlotDefinitionName
from linkml_runtime.utils.compile_python import compile_python
from linkml_runtime.utils.formatutils import underscore
from linkml_runtime.utils.schemaview import SchemaView, ClassDefinitionName
from linkml_runtime.utils.yamlutils import YAMLRoot
from rdflib import Graph

from linkml_datalog.dumpers.tupledumper import TupleDumper
from linkml_datalog.generators.dataloggen import DatalogGenerator
from linkml.utils.datautils import _get_format, infer_root_class, get_loader, dumpers_loaders
from linkml_datalog.model.validation import ValidationReport, ValidationResult


def runcmd(cmd: str) -> int:
    status = os.system(cmd)
    logging.info(f'{status} CMD: {cmd}')
    if status != 0:
        raise Exception(f'Error running" {cmd}')
    return status

@dataclass
class DatalogEngine:
    """
    Engine for performing datalog operations on linkml data

    ALPHA

    uses DatalogDumper

    """
    sv: SchemaView = None
    workdir: str = None

    def run(self, obj: Union[YAMLRoot, Graph], prefix_map: Dict[str, str] = None, strict=True):
        """
        Run datalog inference over a data object
        """
        sv = self.sv
        workdir = self.workdir
        generator = DatalogGenerator(sv.schema)
        with open(os.path.join(workdir, 'schema.dl'), 'w') as stream:
            stream.write(generator.serialize())
        generator.serialize()
        dumper = TupleDumper()
        dumper.dump(obj, sv, directory=workdir, prefix_map=prefix_map)
        result = subprocess.run(['souffle', f'-F{workdir}', f'-D{workdir}', f'{workdir}/schema.dl'],
                                capture_output=True)
        print(f'STDERR: {result.stderr}')
        if result.stderr:
            logging.error(f'STDERR: {result.stderr}')
        if result.stdout:
            logging.error(f'STDOUT: {result.stdout}')
        result.check_returncode()
        if strict and result.stderr:
            raise Exception(f'Got warnings: {result.stderr}')
        #runcmd(f'souffle -F{workdir} -D{workdir} {workdir}/schema.dl')

    def _parse_results(self, pred: str) -> List[List[str]]:
        with open(os.path.join(self.workdir, f'{pred}.csv')) as csvfile:
            reader = csv.reader(csvfile, delimiter='\t', quotechar='|')
            return [row for row in reader]

    def validation_results(self) -> ValidationReport:
        """
        Retrieves validation results, after running souffle
        """
        rows = self._parse_results('validation_result')
        results = []
        for [typ, subject, cls, pred, val, info] in rows:
            result = ValidationResult(type=typ,
                                      subject=subject,
                                      instantiates=cls,
                                      predicate=pred,
                                      object_str=val,
                                      info=info)
            results.append(result)
        return ValidationReport(results=results)

    def inferred_slot_values(self, cn: ClassDefinitionName, sn: SlotDefinitionName) -> List[Tuple[str, str]]:
        return [(r[0], r[1]) for r in self._parse_results(f'{cn}_{sn}')]

    def materialize_inferences(self, obj: YAMLRoot) -> None:
        # TODO: potentially redo, get all inferred triples first
        print(f'MAT {obj} {type(obj)}')
        if obj is None:
            return
        if isinstance(obj, list):
            for x in obj:
                self.materialize_inferences(x)
            return
        if isinstance(obj, dict):
            for x in obj.values():
                self.materialize_inferences(x)
            return
        sv = self.sv
        t = type(obj)
        try:
            cn = t.class_name
        except:
            return
        id_slot = sv.get_identifier_slot(cn)
        if id_slot:
            id_val = getattr(obj, id_slot.name)
        else:
            id_val = None
        for islot in sv.class_induced_slots(cn):
            sn = underscore(islot.name)
            if id_val:
                # TODO: optimize
                for row in self._parse_results(sn):
                    # TODO: CURIE expansion
                    if row[0] == id_val:
                        setattr(obj, sn, row[1])
            if islot.range in sv.all_classes():
                self.materialize_inferences(getattr(obj, sn))










@click.command()
@click.option('--dir', '-d', required=True, help='Directory to export to')
@click.option('--schema', '-s', required=True, help='Path to schema')
@click.option("--input-format", "-f",
              type=click.Choice(list(dumpers_loaders.keys())),
              help="Input format. Inferred from input suffix if not specified")
@click.option("--target-class", "-C",
              help="name of class in datamodel that the root node instantiates")
@click.option("--module", "-m",
              help="Path to python datamodel module")
@click.argument('input')
def cli(input, schema, module, target_class, input_format, dir):
    """
    Performs inference and validation over input files using a linkml schema

    Steps:
     - compile schema to datalog schema (schema.dl)
     - translate instance data to tuples
     - collect above in working directory
     - run souffle
    """
    logging.basicConfig(level=logging.INFO)
    if module is None:
        if schema is None:
            raise Exception('must pass one of module OR schema')
        else:
            python_module = PythonGenerator(schema).compile_module()
    else:
        python_module = compile_python(module)
    sv = SchemaView(schema)
    input_format = _get_format(input, input_format)
    loader = get_loader(input_format)
    if target_class is None:
        target_class = infer_root_class(sv)
    if target_class is None:
        raise Exception(f'target class not specified and could not be inferred')
    py_target_class = python_module.__dict__[target_class]

    obj = loader.load(source=input,  target_class=py_target_class)
    engine = DatalogEngine(sv, workdir=dir)
    engine.run(obj)
    rpt = engine.validation_results()
    print(yaml_dumper.dumps(rpt))


if __name__ == '__main__':
    cli()


# Auto generated from validation.yaml by pythongen.py version: 0.9.0
# Generation date: 2021-12-10T15:22:58
# Schema: reporting
#
# id: https://w3id.org/linkml/reporting
# description: A datamodel for reports on data
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import Nodeidentifier, String
from linkml_runtime.utils.metamodelcore import NodeIdentifier

metamodel_version = "1.7.0"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
OWL = CurieNamespace('owl', 'http://www.w3.org/2002/07/owl#')
PAV = CurieNamespace('pav', 'http://purl.org/pav/')
RDF = CurieNamespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
RDFS = CurieNamespace('rdfs', 'http://www.w3.org/2000/01/rdf-schema#')
REPORTING = CurieNamespace('reporting', 'https://w3id.org/linkml/report')
SCHEMA = CurieNamespace('schema', 'http://schema.org/')
SH = CurieNamespace('sh', 'https://w3id.org/shacl/')
SKOS = CurieNamespace('skos', 'http://www.w3.org/2004/02/skos/core#')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = REPORTING


# Types

# Class references



@dataclass
class ValidationReport(YAMLRoot):
    """
    A report object
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SH.ValidationReport
    class_class_curie: ClassVar[str] = "sh:ValidationReport"
    class_name: ClassVar[str] = "ValidationReport"
    class_model_uri: ClassVar[URIRef] = REPORTING.ValidationReport

    results: Optional[Union[Union[dict, "ValidationResult"], List[Union[dict, "ValidationResult"]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.results, list):
            self.results = [self.results] if self.results is not None else []
        self.results = [v if isinstance(v, ValidationResult) else ValidationResult(**as_dict(v)) for v in self.results]

        super().__post_init__(**kwargs)


@dataclass
class ValidationResult(YAMLRoot):
    """
    An individual result arising from validation of a data instance using a particular rule
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SH.ValidationResult
    class_class_curie: ClassVar[str] = "sh:ValidationResult"
    class_name: ClassVar[str] = "ValidationResult"
    class_model_uri: ClassVar[URIRef] = REPORTING.ValidationResult

    type: Optional[Union[str, NodeIdentifier]] = None
    severity: Optional[Union[str, "SeverityOptions"]] = None
    subject: Optional[Union[str, NodeIdentifier]] = None
    instantiates: Optional[Union[str, NodeIdentifier]] = None
    predicate: Optional[Union[str, NodeIdentifier]] = None
    object: Optional[Union[str, NodeIdentifier]] = None
    object_str: Optional[str] = None
    source: Optional[Union[str, NodeIdentifier]] = None
    info: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.type is not None and not isinstance(self.type, NodeIdentifier):
            self.type = NodeIdentifier(self.type)

        if self.severity is not None and not isinstance(self.severity, SeverityOptions):
            self.severity = SeverityOptions(self.severity)

        if self.subject is not None and not isinstance(self.subject, NodeIdentifier):
            self.subject = NodeIdentifier(self.subject)

        if self.instantiates is not None and not isinstance(self.instantiates, NodeIdentifier):
            self.instantiates = NodeIdentifier(self.instantiates)

        if self.predicate is not None and not isinstance(self.predicate, NodeIdentifier):
            self.predicate = NodeIdentifier(self.predicate)

        if self.object is not None and not isinstance(self.object, NodeIdentifier):
            self.object = NodeIdentifier(self.object)

        if self.object_str is not None and not isinstance(self.object_str, str):
            self.object_str = str(self.object_str)

        if self.source is not None and not isinstance(self.source, NodeIdentifier):
            self.source = NodeIdentifier(self.source)

        if self.info is not None and not isinstance(self.info, str):
            self.info = str(self.info)

        super().__post_init__(**kwargs)


# Enumerations
class ProblemType(EnumDefinitionImpl):

    undeclared_slot = PermissibleValue(text="undeclared_slot",
                                                     description="Applies when a slot is used in data but the slot is undeclared in the datamodel")
    inapplicable_slot = PermissibleValue(text="inapplicable_slot",
                                                         description="Applies when a slot is used in an instance of a class where the slot is not applicable for that class")
    missing_slot_value = PermissibleValue(text="missing_slot_value",
                                                           description="Applies when an instance of a class has a required slot which is not filled in")
    slot_range_violation = PermissibleValue(text="slot_range_violation",
                                                               description="Applies when the value of a slot is inconsistent with the declared range")
    max_count_violation = PermissibleValue(text="max_count_violation",
                                                             meaning=SH.MaxCountConstraintComponent)
    parsing_error = PermissibleValue(text="parsing_error",
                                                 description="The data could not be parsed")

    _defn = EnumDefinition(
        name="ProblemType",
    )

class SeverityOptions(EnumDefinitionImpl):

    FATAL = PermissibleValue(text="FATAL")
    ERROR = PermissibleValue(text="ERROR",
                                 meaning=SH.Violation)
    WARNING = PermissibleValue(text="WARNING",
                                     meaning=SH.Warning)
    INFO = PermissibleValue(text="INFO",
                               meaning=SH.Info)

    _defn = EnumDefinition(
        name="SeverityOptions",
    )

# Slots
class slots:
    pass

slots.type = Slot(uri=SH.sourceConstraintComponent, name="type", curie=SH.curie('sourceConstraintComponent'),
                   model_uri=REPORTING.type, domain=None, range=Optional[Union[str, NodeIdentifier]])

slots.subject = Slot(uri=SH.focusNode, name="subject", curie=SH.curie('focusNode'),
                   model_uri=REPORTING.subject, domain=None, range=Optional[Union[str, NodeIdentifier]])

slots.instantiates = Slot(uri=REPORTING.instantiates, name="instantiates", curie=REPORTING.curie('instantiates'),
                   model_uri=REPORTING.instantiates, domain=None, range=Optional[Union[str, NodeIdentifier]])

slots.predicate = Slot(uri=REPORTING.predicate, name="predicate", curie=REPORTING.curie('predicate'),
                   model_uri=REPORTING.predicate, domain=None, range=Optional[Union[str, NodeIdentifier]])

slots.object = Slot(uri=SH.value, name="object", curie=SH.curie('value'),
                   model_uri=REPORTING.object, domain=None, range=Optional[Union[str, NodeIdentifier]])

slots.object_str = Slot(uri=REPORTING.object_str, name="object_str", curie=REPORTING.curie('object_str'),
                   model_uri=REPORTING.object_str, domain=None, range=Optional[str])

slots.source = Slot(uri=REPORTING.source, name="source", curie=REPORTING.curie('source'),
                   model_uri=REPORTING.source, domain=None, range=Optional[Union[str, NodeIdentifier]])

slots.severity = Slot(uri=REPORTING.severity, name="severity", curie=REPORTING.curie('severity'),
                   model_uri=REPORTING.severity, domain=None, range=Optional[Union[str, "SeverityOptions"]])

slots.info = Slot(uri=REPORTING.info, name="info", curie=REPORTING.curie('info'),
                   model_uri=REPORTING.info, domain=None, range=Optional[str])

slots.validationReport__results = Slot(uri=REPORTING.results, name="validationReport__results", curie=REPORTING.curie('results'),
                   model_uri=REPORTING.validationReport__results, domain=None, range=Optional[Union[Union[dict, ValidationResult], List[Union[dict, ValidationResult]]]])



from dataclasses import dataclass
from datetime import date, datetime

import click
import json
import yaml
import hashlib
from io import StringIO
from contextlib import redirect_stdout
from enum import Enum
from typing import Dict, List, Any, Union, Tuple

DICT_OR_LIST = Union[Dict, List]

class ObjType(Enum):
    LIST = 'l'
    DICT = 'd'
    NUMBER = 'n'
    STRING = 's'
    BOOLEAN = 'b'
    NONE = 'x'
    DOCUMENT = 'r'


HDR = ['id', 'index', 'key', 'val_s', 'val_i', 'type']

@dataclass
class EavDumper:
    """
    Exports arbitrary JSON or Dict objects to EAV-style TSV
    """
    dict_list: List[Dict] = None

    def dumps(self, obj: DICT_OR_LIST, header=False, document_root: Union[dict, str] = None) -> str:
        """
        Dumps a dict to an EAV TSV string

        :param obj:
        :param header:
        :param document_root:
        :return:
        """
        output = StringIO()
        if document_root is not None:
            if isinstance(document_root, str):
                document_root = {
                    '_id': document_root
                }
            document_root['_contents'] = obj
            obj = document_root
        with redirect_stdout(output):
            if header:
                self._header()
            self._dumps(obj)
        return output.getvalue()

    def dump(self, obj: DICT_OR_LIST, file, header=False, **kwargs) -> None:
        """
        Dumps a dict to an EAV TSV file

        :param obj:
        :param header:
        :return:
        """
        with redirect_stdout(file):
            if header:
                self._header()
            self._dumps(obj, **kwargs)

    def as_objs(self, obj: DICT_OR_LIST) -> List[Dict]:
        self.dict_list =[]
        self._dumps(obj)
        r = self.dict_list
        self.dict_list = None
        return r

    def _header(self):
        print("\t".join(HDR))

    def _dumps(self, obj: Any) -> Tuple[Any, ObjType]:
        if isinstance(obj, list):
            id = self._id(obj)
            ix = 0
            for v in obj:
                v2, t = self._dumps(v)
                self.tuple(id, index=ix, key=str(ix), val=v2, type=t)
                ix += 1
            return id, ObjType.LIST
        elif isinstance(obj, dict):
            id = self._id(obj)
            ix = 0
            for k, v in obj.items():
                v2, t = self._dumps(v)
                self.tuple(id, index=ix, key=k, val=v2, type=t)
                ix += 1
            return id, ObjType.DICT
        elif obj is None:
            return obj, ObjType.NONE
        elif isinstance(obj, bool):
            return obj, ObjType.BOOLEAN
        elif isinstance(obj, date) or isinstance(obj, datetime):
            return str(obj), ObjType.STRING
        elif isinstance(obj, float) or isinstance(obj, int):
            return obj, ObjType.NUMBER
        else:
            return str(obj), ObjType.STRING

    def _id(self, obj: DICT_OR_LIST) -> str:
        s = json.dumps(obj, sort_keys=True)
        h = hashlib.md5(str(s).encode('utf-8')).hexdigest()
        # https://datatracker.ietf.org/doc/html/draft-thiemann-hash-urn-01
        return f'urn:hash::md5:{h}'

    def tuple(self, id: str, key: str, index: int, val: Any, type: ObjType):
        if isinstance(val, str):
            val_s = val
            val_i = 0
        elif isinstance(val, bool):
            val_i = int(val)
            val_s = str(val)
        else:
            val_i = val
            val_s = str(val)
        t = [id, index, key, val_s, val_i, type.value]
        t = [str(x).replace("\t","\\t").replace("\n", "\\n") for x in t]
        if self.dict_list is not None:
            self.dict_list.append(dict(id=id, key=key, index=index, val_s=val_s, val_i=val_i, type=type.value))
        else:
            print("\t".join(t))

def add_document_root(obj: Any, path: str) -> Dict:
    document_root = {'_id': path,
                     '_type': 'document',
                     '_contents': obj}
    return document_root


@click.command()
@click.option('--output', '-o', help='Path to output file')
@click.option('--header/--no-header', default=False, help='include a TSV header')
@click.option('--add-root/--no-add-root', default=False, help='include a document root')
@click.option('--type', '-t', help='inject top level type')
@click.argument('input')
def main(input, output, type, add_root: bool, **args):
    with open(input, 'rb') as stream:
        obj = yaml.safe_load(stream)
    if type is not None:
        obj['@type'] = type
    if add_root:
        obj = add_document_root(obj, input)
    dumper = EavDumper()
    if output is None:
        print(dumper.dumps(obj, **args))
    else:
        with open(output, 'w') as stream:
            dumper.dump(obj, stream, **args)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['linkml_datalog',
 'linkml_datalog.dumpers',
 'linkml_datalog.engines',
 'linkml_datalog.generators',
 'linkml_datalog.model',
 'linkml_datalog.utils']

package_data = \
{'': ['*']}

install_requires = \
['linkml>=1.1.13,<2.0.0']

entry_points = \
{'console_scripts': ['linkml-dl = linkml_datalog.engines.datalog_engine:cli']}

setup_kwargs = {
    'name': 'linkml-datalog',
    'version': '0.1.5',
    'description': 'Datalog inference over LinkML schemas and data',
    'long_description': '# linkml-datalog\n\nValidation and inference over LinkML instance data using souffle\n\nDocumentation:\n\n * [linkml.io/linkml-datalog/](https://linkml.io/linkml-datalog/)\n',
    'author': 'cmungall',
    'author_email': 'cjm@berkeleybop.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/linkml/linkml-datalog',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
